/* is_in_viewcone3d
*
* by VEN
*
* This file is provided as is (no warranties).
*/

#include <amxmodx>
#include <fakemeta>
#include <xs>

#define RADIANS_TO_DEGREES(%1) ((%1) * 57.29577951308232)
#define SWITCH_METHOD_VER_ANG_THRESHOLD 30.0
#define VER_HOR_ASPECT_RATIO 0.75

// the geometric concept of this function is that the player's view area can be represented as the pyramid with the infinite height (without real base)
// the aspect ratio of the sides of the imaginary (any) base of the pyramid is 4:3 because usually the aspect ratio of the screen resolution is 4:3 (640:480; 800:600; 1024:768)
// so based on trigonometric concept we can say that Tangent(b) == Tangent(a) * 0.75, so b == ArcTangent(Tangent(a) * 0.75), where a and b is the player's horizontal and vertical FOV/2
// so to check whether the point is belong to that pyramid we can make a condition: player_v_angle - FOV/2 <= point_vector_angle <= player_v_angle + FOV/2 for both horizontal and vertical imaginary planes that lies on the pyramid's height vector
// the tests shows that this method is more accurate than one that engine module use only if -x < player_to_point_vector_vertical_angle < x, where x =~= 30 degrees so for more accuracy the below function uses a combination of the both methods
stock bool:fm_is_in_viewcone3d(index, const Float:point[3]) {
	new Float:v_origin[3], Float:angle_to_point[3]
	pev(index, pev_origin, v_origin)
	pev(index, pev_view_ofs, angle_to_point)
	xs_vec_add(v_origin, angle_to_point, v_origin)
	xs_vec_sub(point, v_origin, angle_to_point)

	new Float:norm[3]
	xs_vec_normalize(angle_to_point, norm)

	engfunc(EngFunc_VecToAngles, angle_to_point, angle_to_point)
	fm_norm_angle(angle_to_point)

	new Float:v_angle[3]
	pev(index, pev_v_angle, v_angle)

	new Float:a
	pev(index, pev_fov, a)
	a *= 0.5

	if (floatabs(angle_to_point[0]) < SWITCH_METHOD_VER_ANG_THRESHOLD) {
		fm_norm_v_angle(v_angle)
		fm_norm_angle(v_angle)

		if (!(v_angle[1] - a <= angle_to_point[1] <= v_angle[1] + a))
			return false

		new Float:b = RADIANS_TO_DEGREES(floatatan(floattan(a, degrees) * VER_HOR_ASPECT_RATIO, _:radian))
		if (!(v_angle[0] - b <= angle_to_point[0] <= v_angle[0] + b))
			return false

		return true
	}

	engfunc(EngFunc_MakeVectors, v_angle)
	global_get(glb_v_forward, v_angle)
	if (xs_vec_dot(norm, v_angle) < floatcos(a * M_PI / 180.0))
		return false

	return true
}

public plugin_init() {
	register_plugin("is_in_viewcone3d", "0.1", "VEN")
	register_clcmd("is_in_viewcone3d", "clcmd_is_in_viewcone3d")
}

public clcmd_is_in_viewcone3d(id) {
	new ent = fm_find_ent_by_owner_(-1, "weaponbox", id)
	if (!ent) {
		server_print("WEAPONBOX NOT FOUND!")
		return PLUGIN_HANDLED
	}

	new Float:origin[3]
	pev(ent, pev_origin, origin)

	new bool:in_view = fm_is_in_viewcone3d(id, origin)

	client_print(id, print_chat, "IN_VIEW[%d]   ENTITY[%d]", in_view, ent)
	client_print(id, print_center, "------- [%d] -------", in_view)
	server_print("IN_VIEW[%d]   ENTITY[%d]", in_view, ent)

	return PLUGIN_HANDLED
}

// standartize v_angle according to VecToAngles angle system
stock fm_norm_v_angle(Float:v_angle[3]) {
	v_angle[0] *= -1.0
	v_angle[0] -= floatround(v_angle[0] / 360.0, floatround_floor) * 360.0
	v_angle[1] -= floatround(v_angle[1] / 360.0, floatround_floor) * 360.0
	if (v_angle[0] < 0.0)
		v_angle[0] += 360.0
	if (v_angle[1] < 0.0)
		v_angle[1] += 360.0
}

// fits angle with range from 0 to 360 in range from -180 to 180
stock fm_norm_angle(Float:angle[3]) {
	if (angle[0] > 180.0)
		angle[0] -= 360.0
	if (angle[1] > 180.0)
		angle[1] -= 360.0
}

stock fm_find_ent_by_owner_(index, const classname[], owner) {
	new ent = index
	while ((ent = engfunc(EngFunc_FindEntityByString, ent, "classname", classname)) && pev(ent, pev_owner) != owner) {}

	return ent
}
